-- WikNukem Forever
-- Copyright (C) 2009 Noé Falzon <firstname.lastname@aliceadsl.fr>

-- This program is  free software:  you can redistribute it and/or modify it
-- under  the terms  of the  GNU General Public License  as published by the
-- Free Software Foundation, version 3 of the License.
--
-- This program  is  distributed  in the hope  that it will  be useful,  but
-- WITHOUT  ANY WARRANTY;  without  even the implied  warranty of MERCHANTA-
-- BILITY  or  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the  GNU General Public License  along
-- with this program. If not, see <http://www.gnu.org/licenses/>.

require "level"

game = {}

levels = {}
levelstack = {}

local playersprite_left = [[
   o 
==v|\
  / \
]]

local playersprite_right = [[
 o   
/|v==
/ \  
]]

playersprite = playersprite_right

local robotsprite = [[
     ,     ,     
    (\____/)     
     (_oo_)      
       (O)       
     __||__    \)
  []/______\[] / 
  / \______/ \/  
 /    /__\       
(\   /____\      
]]

game.init = function()
	
	-- let's be serious a little
	tsize = 10

	-- Get the very first level
	local url
	
	repeat
		url = getRandomURL()
	until url
	
	initLevel(url)
	
	currentmode = 'NORMAL'	-- there is history and compass also

end

initLevel = function(url)
	
	-- check winning condition...
	if url == "/wiki/Duke_Nukem_Forever" then
		setGameMode(win)
	end
	
	-- first check if we're just backtracking
	-- in this case, just pop from stack
	if url == "" then
		if #levelstack > 1 then
			table.remove(levelstack)
			level = levelstack[#levelstack]
		end
		
	-- then check if we've been there already
	else
		local oldlevel = levels[url]
		if oldlevel then
			-- if it's in our current stack, pop all levels after it
			-- ie shortcut if you're looping...
			local ind
			for i,v in ipairs(levelstack) do
				if v == oldlevel then
					ind = i
					break
				end
			end
			
			if ind then
				while #levelstack > ind do
					table.remove(levelstack)
				end
			else
				table.insert(levelstack,oldlevel)
			end
			
			-- in any case, go there
			level = oldlevel	-- ( == levelstack[#levelstack])
			
		-- if we've never been there
		else
			level = makeLevel(url)
			
			-- memorise it in the big list of levels
			levels[url] = level
			table.insert(levelstack,level)
		end
	end
	
	-- have a player
	player = {level.start[1]*tsize, (level.start[2]+1)*tsize/2 - 3}
	pspeed = {0,0}
	jumping = false
	hitpoints = 100
	colliding = false
	firing = false
	weapon = 0	-- 0: machine gun, 1: shotgun
	lastshot = 0
	reloadtime = 1.2 * 1000
	playerdir = 1
	
	-- have a view
	cam = {player[1],player[2]}
	
	levelstart = t or 0
	
	-- reset compass
	comp_target = nil
	comp_scroll = 1
	comp_highlight = 1
	
	
	evol.play(sfx.load)
end

game.update = function(dt,fps,t)

	mode_update[currentmode](dt,fps,t) 
end

game.keydown = function(key)
	
	local H = evol.nh
	
	local quick = evol.isKeyDown(evol.key_lshift)
	if quick then quick = 10
		else quick = 1
	end
	
	if currentmode == 'NORMAL' then
		
		

		if key == evol.key_escape then
			evol.quit()
		
		elseif key == evol.key_space then
			enterExit()
		
		elseif key == evol.key_up then
			pleasejump = true
		
		elseif key == evol.key_f then
			player[2] = player[2] - 1
		
		elseif key == evol.key_tab then
			currentmode = 'HISTORY'
			
		elseif key == evol.key_c then
			currentmode = 'COMPASS'
			
		elseif key == evol.key_left then
			playerdir = -1
			playersprite = playersprite_left
		
		elseif key == evol.key_right then
			playerdir = 1
			playersprite = playersprite_right
		
		elseif key == evol.key_lctrl then
			if weapon == 0 then
				firing = true
			else
				if t > lastshot + reloadtime then
					firing = true
					lastshot = t
					evol.play(sfx.sg)
				end
			end
			
		
		elseif key == evol.key_lshift then
			weapon = (weapon + 1)%2
			
		end
	
	elseif currentmode == 'HISTORY' then
		currentmode = 'NORMAL'
	
	elseif currentmode == 'COMPASS' then
		
		if key == evol.key_escape then
			currentmode = 'NORMAL'
		
		elseif key == evol.key_down then
			comp_highlight = comp_highlight + quick
			if comp_highlight > #level.exits then
				comp_highlight = #level.exits
			end
			if comp_highlight > comp_scroll + H - 5 then
				comp_scroll = comp_scroll + quick
			end

			
		
		elseif key == evol.key_up then
			comp_highlight = comp_highlight - quick
			if comp_highlight < 1 then
				comp_highlight = 1
			end
			if comp_highlight < comp_scroll then
				comp_scroll = comp_highlight
			end
		
				
		elseif key == evol.key_space then
			
			comp_target = comp_highlight
		end
	
	end
	
end

game.keyup = function(key)
	if key == evol.key_lctrl then
		firing = false
	end
end

updateCam = function()
	
	local size = level.size
	-- follow the player, but not on the sides
	
	cam[1] = player[1] - evol.nw/2
	cam[2] = player[2] - evol.nh/2
	
	if cam[1] < 0 then cam[1] = 0 end
	if cam[1] + evol.nw > size*tsize then cam[1] = size*tsize - evol.nw end
	
	if cam[2] < 0 then cam[2] = 0 end
	if cam[2] + evol.nh > size*tsize/2 then cam[2] = size*tsize/2 - evol.nh end
end

updatePlayer = function(dt)
	
	local splim = 70
	local jumpspeed = 35
	local jumpsize = 10
	
	local ladderspeed = 15
	
	local domove = false
	
	-- check if we're on the ground or not (at least one platform)
	-- or ladder
	local pix = checkfloor()
	local ground = false
	local ladder = false
	for i,v in ipairs(pix) do
		ground = ground or v[1] == 'PLATFORM'
		ladder = ladder or v[1] == 'LADDER'
	end
	
	-- x is easy
	
	if evol.isKeyDown(evol.key_right) then
		pspeed[1] = splim
		domove = true
	elseif evol.isKeyDown(evol.key_left) then
		pspeed[1] = -splim
		domove = true
	else
		pspeed[1] = 0
	end
	
	-- y depends on the terrain
	-- First sets of physics: on the ground
	if ground then
		
		if pspeed[2] >= 0 then
			pspeed[2] = 0
		end
		
		if pleasejump then
			jumping = player[2]	-- remember altitude
			pspeed[2] = -jumpspeed
		end
	
	-- Then on the ladder
	elseif ladder then
		
		-- check if we're at the top of the ladder
		-- (ie feet are all in the empty)
		local pix = checkfloor(false,true)	-- check foot level
		local attop = true
		for i,v in ipairs(pix) do
			attop = attop and v[1] == 'EMPTY'
		end
		
		
		if evol.isKeyDown(evol.key_down) then
			pspeed[2] = ladderspeed
			domove = true
		elseif evol.isKeyDown(evol.key_up) then
			-- prevent bouncing when at top
			if not attop then
				pspeed[2] = -ladderspeed
			else
				jumping = player[2]	-- remember altitude
				pspeed[2] = -jumpspeed
			end
			domove = true
		else
			pspeed[2] = 0
		end
		
		-- allow jumping from top of ladder
		if pleasejump and attop then
			jumping = player[2]	-- remember altitude
			pspeed[2] = -jumpspeed
		end
		
	-- Mid air physics: just fall, dumbo!
	else
		if jumping then
			pspeed[2] = -jumpspeed
		else
			pspeed[2] = jumpspeed
		end
		
	end
	
	EC.print(tostring(jumping),0,7)
	
	-- Euler fun again !
	player[1] = player[1] + pspeed[1]*dt
	player[2] = player[2] + pspeed[2]*dt
	
	-- check downward collision (stop at the first solid pix)
	local pix = checkfloor(false,true)	-- check the feet
	local ground = false
	for i,v in ipairs(pix) do
		ground = ground or v[1] == 'PLATFORM'
	end
	
	if ground then player[2] = math.floor(player[2] - 1) end
	
	-- check the end of jumps
	if jumping and player[2] < jumping - jumpsize then
		jumping = false
	end

	-- Borders check
	local w,h = spritesize(playersprite)
	
	if player[1] < 0 then player[1] = 0 end
	if player[1] > level.size*tsize - w then player[1] = level.size*tsize - w end
	
	if player[2] <= 0 then player[2] = 5 ; pspeed[2] = 40 end
	if player[2] > level.size*tsize/2 - h then player[2] = level.size*tsize/2 - h end
	
	
	pleasejump = false
	
	
	-- [[Baddie touching]]
	if colliding then
		hitpoints = hitpoints - 10 * dt
		evol.play(sfx.creepy)
		if hitpoints < 0 then
			hitpoints = 0
			setGameMode(die)
			evol.play(sfx.lose)
		end
	end
	
	if domove and math.random() < 0.5  then
		evol.play(sfx.step)
	end
	
end


checkfloor = function(draw, feet)
	
	local pix = {}
	
	local offset = 0
	if feet then offset = -1 end
	
	
	-- Check status of the pixels below us
	local w,h = spritesize(playersprite)
	local px,py = math.floor(player[1]), math.floor(player[2])
	
	
	for x = px,px + w - 1 do
		local y = py + h + offset
		
		local tx,ty = math.floor(x/tsize),math.floor(y/tsize*2)
		
		local tile = level.map[tx + ty * level.size]
		
		if tile == nil then tile = 'PLATFORM' end	-- careful with the bottom row
		
		table.insert(pix, {tile, x, y})
	end
	
	if draw then
		local str = ""
		for i,v in ipairs(pix) do
			EC.setcol(v[2]-cam[1],v[3]-cam[2],-1,0,0,0,150,0)
			str = str .. v[1] .. ","
		end
		
		EC.print(str, 0,5)
	end
	
	return pix
end


enterExit = function()
	
	local px,py = math.floor(player[1]), math.floor(player[2])
	local tx,ty = math.floor(px/tsize),math.floor(py/tsize*2)
	
	local tile = level.map[tx + ty * level.size]
	
	if type(tile) == 'number' then
	
		local url = level.exits[tile][1]
		initLevel(url)
	end
end



mode_update = {}

mode_update['NORMAL'] = function(dt,fps,t)
	
	EC.fill(nil, " ", nil, {40,40,40})
	
	--[[UPDATING]]
	colliding = false
	for i,v in ipairs(level.baddies) do
		updateBaddie(v,dt)
	end

	
	-- check the dead baddies
	local i = 1
	while i <= #level.baddies do
		if level.baddies[i].hitpoints <= 0 then
			table.remove(level.baddies,i)
			evol.play(sfx.robodie)
		else
			i = i + 1
		end
	end
	
	if t > levelstart + 500 then
		updatePlayer(dt)
	end
	

	
	updateCam(dt)
	
	--[[ DRAWING ]]
	
	-- level
	drawLevel()
	
	if t > levelstart + 500 then
		-- debug + ugly hack
		local ground = checkfloor(false)
		local x,y = ground[1][2],ground[1][3]-3

		-- player
		local sprite
		if playerdir == 1 then
			sprite = playersprite_right
		else
			sprite = playersprite_left
		end
		blit(sprite,x-cam[1],y-cam[2],nil,{0,0,200},nil," ")
		
		-- fire
		local blasts = {
			{0,0},
			{1,0},
			{2,0},
			{3,0},
			{4,0},
			{1,1},
			{1,-1},
		}
		
		if firing then
			
			if math.random() < 0.7 then evol.play(sfx.mg) end
			
			
			for i,v in ipairs(blasts) do
				local patt = "-:.~^'*    "
				local ind = math.random(#patt)
				local char = patt:sub(ind,ind)
				
				local x2
				if playerdir == 1 then
					x2 = x - cam[1] + 5 + v[1]
				else
					x2 = x - cam[1] - 1 - v[1]
				end
				local y2 = y - cam[2] + 1 + v[2]
				
				EC.addch(char,x2,y2)
				EC.setcol(x2,y2,math.random(0,255),math.random(0,255),0)
			end
		end
	end
	
	-- baddies
	drawBaddies()
	
	if weapon == 1 then	-- yeah, shot gun is one hit only
		firing = false
	end
	
	local str = "Hitpoints: " .. math.floor(hitpoints)
	EC.print(str, 0,0)
	EC.fill({0,0,#str,1}, nil, {255,255,255})
	
	if weapon == 0 then str = "Machinegun"
		else str = "Shotgun" end
	EC.print(str, 0,1)
	EC.fill({0,1,#str,1}, nil, {255,255,255})
	
	-- Level title
	local glow = math.sin(2*math.pi*t/500)/8 + 0.125
	glow = glow*255
	EC.print(level.title, 0,evol.nh-1)
	EC.fill({0,evol.nh-1,#level.title,1},nil,{0,255,0},{glow,glow,glow})
	
	-- Compass thingy
	if comp_target then
		local dist = (player[1] - level.exits[comp_target][3]*tsize)^2
		dist = dist + (player[2] - level.exits[comp_target][4]*tsize/2)^2
		dist = math.ceil(math.sqrt(dist))
		
		local str = "Target distance = " .. dist
		EC.print(str, evol.nw - #str,evol.nh-1)
		EC.fill({evol.nw - #str,evol.nh-1,#str,1},nil,{0,255,0},{glow,glow,glow})
	end
	EC.refresh()
end


mode_update['HISTORY'] = function()
	
	EC.fill(nil, " ",{0,250,0},{0,0,0})
	
	EC.print("History", evol.nw/2 - #("History")/2)
	
	for i,v in ipairs(levelstack) do
		EC.print(v.title, evol.nw/2 - #v.title/2,i+1)
	end
	
	EC.refresh()
end

mode_update['COMPASS'] = function()
	
	EC.fill(nil, " ",{0,250,0},{0,0,0})
	
	local H = evol.nh
	
	if #level.exits < H-4 then
		
		for i,v in ipairs(level.exits) do
			local posx = evol.nw/2 - #v[2]/2
			
			EC.print(v[2],posx,i + 1)
			if i == comp_target then
				EC.fill({posx,i + 1,#v[2],1},nil,{0,0,255})
			end
			
			if i == comp_highlight then
				EC.fill({posx,i + 1,#v[2],1},nil,nil, {100,100,255})
			end
			
		end

	else
		
		for i,v in ipairs(level.exits) do
			
			local posx = evol.nw/2 - #v[2]/2
			
			if i >= comp_scroll and i < comp_scroll + H -4 then
				local ypos = i - comp_scroll + 2
				
				EC.print(v[2],posx,ypos)
				
				if i == comp_target then
					EC.fill({posx,ypos,#v[2],1},nil,{0,0,255})
				end
				
				if i == comp_highlight then
					EC.fill({posx,ypos,#v[2],1},nil,nil, {100,100,255})
				end
			end
			
			if comp_scroll > 1 then
				EC.print("^ more ^",0,0)
			end
			if #level.exits > comp_scroll + H - 5 then
				EC.print("v more v",0,H-1)
			end
		end
	end
	
	EC.refresh()
end


updateBaddie = function(v,dt)
	
	local speed = 20
	if v[5] == 0 then
		speed = -speed
	end
	
	v[1] = v[1] + speed*dt
	
	if v[1] < v[3] then
		v[1] = v[3]
		v[5] = (v[5] + 1)%2
	elseif v[1] > v[4] then
		v[1] = v[4]
		v[5] = (v[5] + 1)%2
	end
	
	-- check collision with player
	local pw,ph = spritesize(playersprite)
	local bw,bh = spritesize(robotsprite)
	local px, py = player[1] + pw/2, player[2] + ph/2
	local bx, by = v[1] + bw/2, v[2] + bh/2 - bh	-- remember their true positions are in the ground
	
	if math.abs(px - bx) < (pw + bw)/2 and
		math.abs(py - by) < (ph + bh)/2 then
		
		colliding = true
	end
	
	-- check if dude is shooting our ass!
	-- (vague alignment is required)
	
	-- O: machine gun
	local range = 50
	local damage = 60
	
	-- 1: shotgun
	if weapon == 1 then
		range = 25
		damage = 50
	end
	
	v.hit = false
	v.hitpoints = v.hitpoints or 100
	
	if firing then
		if math.abs(px - bx) < range and (px - bx)*playerdir < 0 and
			math.abs(py - by) < (ph + bh)/2 then
			
			if math.random() < 0.7 then evol.play(sfx.simple) end
			
			if weapon == 0 then
				v.hitpoints = v.hitpoints - damage*dt
				v.hit = true
			else
				v.hitpoints = v.hitpoints - damage
				v.hit = true
			end
		end
	end
end

drawBaddies = function()
	for i,v in ipairs(level.baddies) do
		local bx,by = v[1],v[2]
		local w,h = spritesize(robotsprite)
		
		if math.abs(bx-player[1]) < evol.nw + 20 and
			math.abs(by-player[2]) < evol.nh + 20 then
			
			local col = {255,0,0}
			if v.hit then
				col = {math.random(0,255),0,0}
				
				for n = 1,10 do
					local bulx,buly = math.random(bx,bx+w-1)-cam[1],math.random(by,by+h-1)-cam[2]-h
					local patt = "-:.~^'*"
					local ind = math.random(#patt)
					local char = patt:sub(ind,ind)
					
					EC.addch(char,bulx,buly)
					EC.setcol(bulx,buly,math.random(0,255),math.random(0,255),0)
				end
				
			end
			
			
			blit(robotsprite,v[1]-cam[1],v[2]-cam[2]-h, false,col,nil," ")
			--EC.print(tostring(v.hitpoints),v[1]-cam[1],v[2]-cam[2]-h)
		end
	end
end
