-- WikNukem Forever
-- Copyright (C) 2009 Noé Falzon <firstname.lastname@aliceadsl.fr>

-- This program is  free software:  you can redistribute it and/or modify it
-- under  the terms  of the  GNU General Public License  as published by the
-- Free Software Foundation, version 3 of the License.
--
-- This program  is  distributed  in the hope  that it will  be useful,  but
-- WITHOUT  ANY WARRANTY;  without  even the implied  warranty of MERCHANTA-
-- BILITY  or  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the  GNU General Public License  along
-- with this program. If not, see <http://www.gnu.org/licenses/>.

local http = require "socket.http"

local laddersprite = [[
H--------H
H--------H
H--------H
H--------H
H--------H
]]

local platsprite = [[
##########
#.#.#.#.#.
.x.x.x.x.x
.#.#.#.#.#
##########
]]

local exitsprite = [[
   ____   
  /0000\  
 /0    0\ 
 |0    0| 
 |0    0| 
]]


local function WP(url)
	return "http://en.wikipedia.org/" .. url
end

function getRandomURL()
	
	local text,code,headers = http.request("http://en.wikipedia.org/wiki/Special:Random")
	
	-- find title and URL (needed when getting random page)
	local _,_,realurl = text:find('<a href="([^%s:"%?]-)" title="View the content')
	
	return realurl
	
end

local function getLinksFromURL(url)
	
	local text,code,headers = http.request(url)
	
	local links = {}
	
	-- find title
	local _,_,title = text:find('<title>(.*) %- Wikipedia, the free encyclopedia</title>')
	
	-- get all links
	-- the grep pattern should get only links to other "valid" pages
	
	for link,label in text:gmatch('<a href="([^%s:"%?]-)" title="([^"]-)">') do
		links[link] = label
	end
	
	return title,links
end

-- Map format: strings for normal tiles,
-- numbers for exits (the index in the exits table)
-- Also, they are always square.
makeLevel = function(url)

	local level = {}
	
	local title, links = getLinksFromURL(WP(url))
	
	-- counting links
	local Nlinks = 0
	for k,v in pairs(links) do
		Nlinks = Nlinks + 1
	end
	
	
	-- deciding the size of the level
	-- about 2 or 3 links per screen
	-- tiles are about 16 chars wide
	local size = math.ceil(math.sqrt(Nlinks/2.5) * evol.nw / tsize)
	
	EC.print("Fetched " .. url .. ":" .. size,0,0)
	EC.refresh()
	
	-- make empty map
	local map = {}
	for i=0,size^2-1 do
		map[i] = "EMPTY"
	end
	
	-- retrieve links and place them randomly on the map
	-- and ensure no two links are on the same spot...
	local exits = {}
	local indexes = {}
	for i = 0,size^2-1 do
		table.insert(indexes,i)
	end
	
	
	for k,v in pairs(links) do
		local index = table.remove(indexes,math.random(#indexes))
		
		-- make a random color
		local col = hsv2rgb(math.random(0,359),1,1)
		
		table.insert(exits,{k,v, math.floor(index%size), math.floor(index/size), col})
	end
	
	-- add a starting point/exit
	local index = table.remove(indexes,math.random(#indexes))
	local start_exit = {math.floor(index%size), math.floor(index/size)}
	-- make a random color
	local col = hsv2rgb(math.random(0,359),1,1)
	
	table.insert(exits,{"","Exit", start_exit[1], start_exit[2], col})

	
	-- cross reference the links from the map
	for i,v in pairs(exits) do
		map[v[3] + v[4]*size] = i
	end
	
	-- make platforms
	for i,v in ipairs(exits) do
		local len = math.random(2,size/2)
		local shift = math.random(0,len-1)
		local y = v[4]+1
		
		-- spawn them from the tile under the exit, to be sure it's reached
			-- to the right
		for x = v[3],v[3]-shift+len-1 do
			if x >= 0 and x < size and y >= 0 and y < size then
				if map[x + y*size] == 'EMPTY' then
					map[x + y*size] = 'PLATFORM'
				else
					break
				end
			end
		end
			-- to the left
		for x = v[3]-1,v[3]-shift,-1 do
			if x >= 0 and x < size and y >= 0 and y < size then
				if map[x + y*size] == 'EMPTY' then
					map[x + y*size] = 'PLATFORM'
				else
					break
				end
			end
		end
			
	end
	
	-- remove double thick platforms (the top one)
	for i = 0,size^2-1 do
		if map[i] == 'PLATFORM' and map[i+size] == 'PLATFORM' then
			map[i] = 'EMPTY'
		end
	end
	
	-- make ladders
	-- pick a random column, check all platform elements there,
	-- pick two of those at random, draw a ladder between.
	-- if there is no platform, forget it, if only one, make a ladder
	-- to the ground
	
	for i = 1,#exits do
		local col = math.random(0,size-1)
		local plats = {}
		
		for y = col,size^2,size do
			if map[y] == 'PLATFORM' then
				table.insert(plats,y)
			end
		end
		
		if #plats == 1 then
			for y = plats[1],size^2,size do
				if type(map[y]) ~= 'number' then
					map[y] = 'LADDER'
				end
			end
		elseif #plats > 1 then
			local ind = math.random(#plats-1)
			local start = plats[ind]
			local ending = plats[ind+1]
			
			for y = start,ending,size do
				if type(map[y]) ~= 'number' then
					map[y] = 'LADDER'
				end
			end
		
		end
		
	
	end
	
	-- remove double thick ladders (the right one)
	for i = 0,size^2-1 do
		if map[i] == 'LADDER' and map[i+1] == 'LADDER' then
			map[i+1] = 'EMPTY'
		end
	end

	
	--[[ BADDIES !! ]]
	
	baddies = {}
	for i = 1, #exits do
		
		local x,y = math.random(0, size-1), math.random(0,size-1)
		
		-- go down to the closest floor (or the bottom row)
		while map[x + y*size] ~= 'PLATFORM' and map[x + y*size] ~= nil do
			y = y+1
		end
		
		-- compute range of roaming
		local minx, maxx = x,x
		while minx > 0 and (map[minx-1 + y*size] == 'PLATFORM' or map[minx-1 + y*size] == 'LADDER') do
			minx = minx - 1
		end
		while maxx < size-1 and (map[maxx+1 + y*size] == 'PLATFORM' or map[maxx+1 + y*size] == 'LADDER') do
			maxx = maxx + 1
		end
		
			-- special case
		if y == size then
			minx,maxx = 0,size-1
		end
		
		-- random direction
		local dir = math.random(0,1)
		
		table.insert(baddies, {x*tsize,y*tsize/2,minx*tsize,maxx*tsize,dir})
	end
	
	
	
	-- pack it up
	level.title = title
	level.size = size
	level.map = map
	level.exits = exits
	level.start = start_exit
	level.baddies = baddies
	
	return level
end

drawTile = function(i,j)
	
	if i < 0 or i >= level.size or j < 0 or j >= level.size then
		return
	end
	
	local tile = level.map[i + j*level.size]
	
	local x,y = i*tsize - cam[1], j*tsize/2 - cam[2]
	
	if tile == 'EMPTY' then
		--EC.fill({x,y,tsize,tsize/2},nil,nil,{255,255,200})
	elseif tile == 'LADDER' then
		blit(laddersprite,x,y,nil,{100,100,100})
	elseif tile == 'PLATFORM' then
		blit(platsprite,x,y,nil,{153,128,0})

		
	elseif type(tile) == 'string' then
		EC.fill({x,y,tsize,tsize/2},nil,nil,{((i+j)*37)%256,0,140})
		EC.print(tile,x,y)
	-- exits
	else	
		blit(exitsprite,x,y,nil,{200,200,200})
		return level.exits[tile]
	end
end

drawTileAnyway = function(i,j)

	if i < 0 or i >= level.size or j < 0 or j >= level.size then
		return
	end
	
	local x,y = i*tsize - cam[1], j*tsize/2 - cam[2]
	
	EC.fill({x,y,tsize,tsize/2},nil,nil,{255,240,0})
end

drawLevel = function()

	EC.fill(nil," ",{255,255,255},{0,0,0})
	
	-- check which tiles are on the screen
	-- start at topleft and add the number there is in the screen
	local tx,ty = math.floor(cam[1]/tsize), math.floor(cam[2]/(tsize/2))
	if tx < 0 then tx = 0 end
	if ty < 0 then ty = 0 end
	
	-- well, draw them
	
	local exits = {}
	for i = tx,tx+math.ceil(evol.nw/tsize) do
		for j = ty,ty+math.ceil(evol.nw/tsize*2) do
			local tmp = drawTile(i,j)	-- will return an exit if it's one
			if tmp then table.insert(exits,tmp) end
		end
	end
	
	-- write the exits names last
	
	local colmap = {}
	
	for i,ex in ipairs(exits) do
		local x,y = ex[3]*tsize - cam[1], ex[4]*tsize/2 - cam[2]
		local str = ex[2]

		EC.print(str, x - #str/2 + tsize/2, y)
		EC.fill({x - #str/2 + tsize/2, y, #str,1}, nil, {200,200,200})
		
		-- maybe some glowing...
		local span = 10
		local cx = x + tsize/2
		local cy = y + tsize/4
		

		for u = cx - span,cx + span do
			
			for v = cy - span/2,cy + span/2 do
				
				if u>=0 and u<evol.nw then
					
					local r,g,b = ex[5][1],ex[5][2],ex[5][3]
					local dist = math.exp(-math.sqrt((u-cx)^2 + ((v-cy)*2)^2)/3)
					
					local phase = str:byte(1)
					dist = dist * (math.sin(2*math.pi*t/2000 + phase)/2 + 0.5)
					
					-- get color
					local old = colmap[math.floor(u) + math.floor(v)*evol.nw] or {0,0,0}
					local new = {r*dist,g*dist,b*dist}
					colmap[math.floor(u) + math.floor(v)*evol.nw] = {old[1]+new[1],old[2]+new[2],old[3]+new[3]}
				end
			end
		end
		
	end
	
	-- apply our beautiful colors
	for k,v in pairs(colmap) do
		local r,g,b = v[1],v[2],v[3]
		
		if v[1] > 255 then v[1] = 255 end
		if v[2] > 255 then v[2] = 255 end
		if v[3] > 255 then v[3] = 255 end
		if v[1] < 0 then v[1] = 0 end
		if v[2] < 0 then v[2] = 0 end
		if v[3] < 0 then v[3] = 0 end
		
		
		local x,y = math.floor(k%evol.nw), math.floor(k/evol.nw)
		EC.setcol(x,y,nil,nil,nil,r,g,b)
	end 

end
