-- WikNukem Forever
-- Copyright (C) 2009 Noé Falzon <firstname.lastname@aliceadsl.fr>

-- This program is  free software:  you can redistribute it and/or modify it
-- under  the terms  of the  GNU General Public License  as published by the
-- Free Software Foundation, version 3 of the License.
--
-- This program  is  distributed  in the hope  that it will  be useful,  but
-- WITHOUT  ANY WARRANTY;  without  even the implied  warranty of MERCHANTA-
-- BILITY  or  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the  GNU General Public License  along
-- with this program. If not, see <http://www.gnu.org/licenses/>.

local EC = evol.curses

local splash = [[
----------------------------------
  ______                      __  
 /_  __/__  ____  ____  _____/ /_ 
  / / / _ \/ __ \/ __ \/ ___/ __ \
 / / /  __/ / / / /_/ / /__/ / / /
/_/  \___/_/ /_/\____/\___/_/ /_/ 
           Productions            
----------------------------------
]]

local title= [[
 __    __ _ _        __       _                  
/ / /\ \ (_) | __ /\ \ \_   _| | _____ _ __ ___  
\ \/  \/ / | |/ //  \/ / | | | |/ / _ \ '_ ` _ \ 
 \  /\  /| |   </ /\  /| |_| |   <  __/ | | | | |
  \/  \/ |_|_|\_\_\ \/  \__,_|_|\_\___|_| |_| |_|
                                                 
          ___                                    
         / __\__  _ __ _____   _____ _ __        
        / _\/ _ \| '__/ _ \ \ / / _ \ '__|       
       / / | (_) | | |  __/\ V /  __/ |          
       \/   \___/|_|  \___| \_/ \___|_|          
]]

local doFadingEffet = function()
	
	EC.fill({0,0,evol.nw,evol.nh},nil,nil,{0,0,0})
	
	local pix = {}
	
	for i=1,evol.nw*evol.nh do
		table.insert(pix,i-1)
	end
	
	while #pix > 0 do
		local i = math.random(#pix)	
			
		EC.addch(" ", pix[i] % evol.nw, pix[i] / evol.nw)
	
		if #pix % 50 == 0 then
			EC.refresh()
			evol.flip()
		end
		
		table.remove(pix,i)
	end
	
	EC.print("Fini: " .. #pix,0,0)
	EC.refresh()
end



menu = {}

menu.init = function()
	
	issplash = true
	evol.play(sfx.tidi)
end


menu.update = function(dt,fps,t)
	if issplash then
		menu.splash(dt,fps,t)
	else
		menu.menu(dt,fps,t)
	end
end


menu.splash = function(dt,fps,t)
	
	local c,wlen = 24/1000, 20
	local W,H = evol.nw, evol.nh
	
	local cenx, ceny = W/2, H/2
	
	local pixrat = evol.cw/evol.ch
	
	for i = 0,W-1 do
		for j = 0,H-1 do
			local dist = math.sqrt((i-cenx)^2 + ((j-ceny)/pixrat)^2)
			local amp = math.sin(2*math.pi/wlen*(dist - c*t))
			
			EC.addch( "@" ,i,j)
			EC.setcol(i,j,255*(1-amp)/2,255*(1-amp)/2,255)
		end
	end
	
	blit(splash, cenx,ceny, true, {0,184,138})
		
	EC.refresh()
end



menu.keydown = function(key)
	
	if issplash then
		issplash = false
		evol.play(sfx.intro)
	else
		evol.play(sfx.long)
		doFadingEffet()
		setGameMode(game)
	end
	
	if key == evol.key_escape then
		evol.quit()
	end
	
end

local pattern = ".:;'~\"`´"

menu.menu = function(dt,fps,t)
	local W,H = evol.nw, evol.nh
	
	for i = 0,W-1 do
		for j = 0,H-1 do
			local amp = math.sin(2*math.pi/30*(i - 10/1000*t))
			amp = amp + math.sin(2*math.pi/50*(j - 15/1000*t))
			amp = amp + math.sin(2*math.pi/40*(i+j - 30/1000*t))
			amp = amp + math.sin(2*math.pi/45*(j-i - 20/1000*t))
					
			amp = amp/4 + 1
			amp = amp/2	-- should be 0..1 now
			
			EC.addch(pattern:sub(math.random(#pattern)),i,j)
			--EC.setcol(i,j,255,255*(amp-1)/2,0, 0,0,0)
			EC.setcol(i,j,255,255*amp,0, 0,0,0)
		end
	end
	
	blit(title, W/2, H/2, true, {255,255,0},{math.random(0,255),0,0})
	
	EC.print("Press a key to start", 30, H-2)
	
	EC.refresh()
end
